package org.siqisource.agave.test;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.dbunit.Assertion;
import org.dbunit.DatabaseUnitException;
import org.dbunit.database.DatabaseConnection;
import org.dbunit.dataset.Column;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.DefaultDataSet;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.ITable;
import org.dbunit.dataset.excel.XlsDataSet;
import org.dbunit.operation.DatabaseOperation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.core.io.DefaultResourceLoader;
import org.springframework.test.context.TestContext;
import org.springframework.test.context.support.AbstractTestExecutionListener;

public class XlsDataSetTestExecutionListener extends AbstractTestExecutionListener {

	protected final Logger log = LoggerFactory.getLogger(getClass());

	public void beforeTestMethod(TestContext testContext) throws Exception {
		Class<?> testClass = testContext.getTestClass();
		Method testMethod = testContext.getTestMethod();

		DataSet annotation = (DataSet) AnnotationUtils.findAnnotation(testMethod, DataSet.class);
		if (annotation == null) {
			return;
		}

		final InputStream ins = getFileInputStream(testClass, annotation);
		DatabaseConnection databaseConnection = getDatabaseConnection(testContext, annotation);

		IDataSet dataSet = new XlsDataSet(ins);

		this.initData(annotation, testContext, dataSet);

		String[] tables = annotation.excelTabs();
		for (int i = 0; i < tables.length; i++) {
			String tablePairs = tables[i];
			ITable preTable = dataSet.getTable(tablePairs.split(",")[0]);
			DefaultDataSet dds = new DefaultDataSet(preTable);
			DatabaseOperation.CLEAN_INSERT.execute(databaseConnection, dds);
		}
		ins.close();
		databaseConnection.close();
	}

	private void initData(DataSet annotation, TestContext testContext, IDataSet dataSet) throws DataSetException {
		String listData = annotation.listData();
		if (listData != null && !"".equals(listData) && dataSet.getTable(listData).getRowCount() > 0) {
			this.setOperateData(testContext, dataSet.getTable(listData));
		}

		String mapData = annotation.mapData();
		if (mapData != null && !"".equals(mapData) && dataSet.getTable(mapData).getRowCount() > 0) {
			this.setPartitiveFields(testContext, dataSet.getTable(mapData));
		}

		String expressionData = annotation.expressionData();
		if (expressionData != null && !"".equals(expressionData) && dataSet.getTable(expressionData).getRowCount() > 0) {
			this.setExpressionData(testContext, dataSet.getTable(expressionData));
		}
	}

	public void afterTestMethod(TestContext testContext) throws Exception {
		Class<?> testClass = testContext.getTestClass();
		Method testMethod = testContext.getTestMethod();
		DataSet annotation = (DataSet) AnnotationUtils.findAnnotation(testMethod, DataSet.class);
		if (annotation == null) {
			return;
		}
		final InputStream ins = getFileInputStream(testClass, annotation);
		DatabaseConnection databaseConnection = getDatabaseConnection(testContext, annotation);

		IDataSet dataSet = new XlsDataSet(ins);
		String[] tables = annotation.excelTabs();

		IDataSet databaseDataSet = databaseConnection.createDataSet();
		for (int i = 0; i < tables.length; i++) {
			String tablePairs = tables[i];
			if (tablePairs.indexOf(',') == -1) {
				continue;
			}
			String[] tableNames = tablePairs.split(",");
			ITable actualTable = databaseDataSet.getTable(tableNames[0].trim());
			ITable expectedTable = dataSet.getTable(tableNames[1].trim());
			if (tableNames.length > 2) {
				ITable ignoreTable = dataSet.getTable(tableNames[2].trim());
				String[] ignoreCols = this.getIgnoreCols(ignoreTable);
				Assertion.assertEqualsIgnoreCols(expectedTable, actualTable, ignoreCols);
			} else {
				Assertion.assertEquals(expectedTable, actualTable);
			}
		}
		ins.close();
		databaseConnection.close();
	}

	private String[] getIgnoreCols(ITable ignoreTable) {
		String[] ignoreCols = null;
		try {
			Column[] columns = ignoreTable.getTableMetaData().getColumns();
			ignoreCols = new String[columns.length];
			for (int i = 0; i < columns.length; i++) {
				ignoreCols[i] = columns[i].getColumnName();
			}
		} catch (DataSetException e) {
			e.printStackTrace();
		}
		return ignoreCols;
	}

	public static String getPackageName(Class<?> clazz) {
		String className = clazz.getName();
		int lastDotIndex = className.lastIndexOf(".");
		return (lastDotIndex != -1 ? className.substring(0, lastDotIndex + 1) : "");
	}

	public static InputStream getFileInputStream(Class<?> testClass, DataSet annotation) throws IOException {
		String filePath = "classpath:/" + getPackageName(testClass).replace('.', '/');
		String file = annotation.file();
		if ("".equals(file)) {
			file = testClass.getSimpleName();
		}
		String fileLocation = filePath + file;
		return new DefaultResourceLoader().getResource(fileLocation).getInputStream();

	}

	public static DatabaseConnection getDatabaseConnection(TestContext testContext, DataSet annotation) throws SQLException, DatabaseUnitException {
		DataSource datasource = (DataSource) testContext.getApplicationContext().getBean(annotation.dataSource());
		Connection connection = datasource.getConnection();
		return new DatabaseConnection(connection);
	}

	private void setExpressionData(TestContext testContext, ITable dataSet) throws DataSetException {
		try {
			Class<?> operateDataClazz = this.getOperateDataClazz(testContext);
			List<Expression> expressions = new ArrayList<Expression>();
			Expression expression;
			Object propertyName = "";
			String type = "";
			Object value = "";
			for (int i = 0, rows = dataSet.getRowCount(); i < rows; i++) {
				expression = new Expression();
				propertyName = dataSet.getValue(i, "propertyName");
				value = dataSet.getValue(i, "value");
				if (propertyName != null && value != null) {
					expression.setPrefix(Converter.objectToString(dataSet.getValue(i, "prefix")));
					expression.setPropertyName(Converter.objectToString(propertyName));
					expression.setCompareSymbol(Converter.objectToString(dataSet.getValue(i, "compareSymbol")));
					type = operateDataClazz.getDeclaredField(propertyName.toString()).getType().getName();
					expression.setValue(Converter.valueTransform(type, value.toString()));
					expression.setSuffix(Converter.objectToString(dataSet.getValue(i, "suffix")));
					expressions.add(expression);
				}
			}
			Field expressionDatas = testContext.getTestClass().getDeclaredField("expressions");
			expressionDatas.setAccessible(true);
			expressionDatas.set(testContext.getTestInstance(), expressions);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void setOperateData(TestContext testContext, ITable dataSet) {
		try {
			Class<?> operateDataClazz = this.getOperateDataClazz(testContext);

			// 获取实体中定义的字段及字段类型
			Column[] columns = dataSet.getTableMetaData().getColumns();
			Map<String, Class<?>> columnTypes = getColumnTypes(operateDataClazz, columns);

			List<Object> operateData = new ArrayList<Object>();
			String columnName = "";
			String attribute = "";
			Class<?> type = null;
			Object data = null;
			for (int i = 0, rows = dataSet.getRowCount(); i < rows; i++) {
				Object entity = operateDataClazz.newInstance();
				for (Column column : columns) {
					columnName = column.getColumnName();
					attribute = Converter.getFirstUpperProperty(columnName);
					type = columnTypes.get(columnName);
					data = dataSet.getValue(i, columnName);
					if (data != null) {
						Method method = entity.getClass().getMethod("set" + attribute, type);
						method.invoke(entity, Converter.valueTransform(type.getName(), data.toString()));
					}
				}
				operateData.add(entity);
			}
			Field operateDatas = testContext.getTestClass().getDeclaredField("listData");
			operateDatas.setAccessible(true);
			operateDatas.set(testContext.getTestInstance(), operateData);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void setPartitiveFields(TestContext testContext, ITable dataSet) {
		try {
			Class<?> operateDataClazz = this.getOperateDataClazz(testContext);

			// 获取PartitiveFields的类别
			Field fieldDatas = testContext.getTestClass().getDeclaredField("mapData");
			Class<?> fieldClazz = (Class<?>) fieldDatas.getGenericType();

			// 获取实体中定义的字段及字段类型
			Column[] columns = dataSet.getTableMetaData().getColumns();
			Map<String, Class<?>> columnTypes = getColumnTypes(operateDataClazz, columns);

			String columnName = "";
			Object entity = fieldClazz.newInstance();
			Object data = null;
			String type = "";
			for (Column column : columns) {
				columnName = column.getColumnName();
				data = dataSet.getValue(0, columnName);
				try {
					type = columnTypes.get(columnName).getName();
				} catch (Exception e) {
					type = "java.lang.String";
				}
				if (data != null) {
					Method method = fieldClazz.getMethod("put", Object.class, Object.class);
					method.invoke(entity, Converter.columnToProperty(columnName), Converter.valueTransform(type, data.toString()));
				}
			}
			fieldDatas.setAccessible(true);
			fieldDatas.set(testContext.getTestInstance(), entity);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private Class<?> getOperateDataClazz(TestContext testContext) {
		Class<?> operateDataClazz = null;
		try {
			// 获取操作数据对应的实体类别
			Field operateDatas = testContext.getTestClass().getDeclaredField("listData");
			ParameterizedType parameterizedType = (ParameterizedType) operateDatas.getGenericType();
			operateDataClazz = (Class<?>) parameterizedType.getActualTypeArguments()[0];
		} catch (Exception e) {
			e.printStackTrace();
		}
		return operateDataClazz;
	}

	private Map<String, Class<?>> getColumnTypes(Class<?> entity, Column[] columns) {
		Map<String, Class<?>> columnTypes = new HashMap<String, Class<?>>();
		try {
			String columnName = "";
			for (Column column : columns) {
				columnName = column.getColumnName();
				columnTypes.put(columnName, entity.getDeclaredField(Converter.columnToProperty(columnName)).getType());
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return columnTypes;
	}
}
